import json

import grpc
from sentinel_protobuf.sentinel.node.v2.msg_pb2 import MsgSubscribeRequest
from sentinel_protobuf.sentinel.session.v2.msg_pb2 import MsgEndRequest, MsgStartRequest

from sentinel_sdk.sdk import SDKInstance
from sentinel_sdk.types import NodeType, TxParams, Status
from sentinel_sdk.utils import search_attribute

# GRPC_VERBOSITY=debug GRPC_TRACE=tcp,http python test/queryall.py

# NEVER SHARE YOUR SECRET!
secret = input("your secret: ")

sdk = SDKInstance(grpcaddr="grpc.sentinel.co", grpcport=9090, secret=secret)

# The address can be founded in every modules because is inherited
# For example: sdk.sessions._account.address
print(sdk.nodes._account.address)

node_address = input("give a node address: ")
node = sdk.nodes.QueryNode(node_address)
node_status = json.loads(sdk.nodes.QueryNodeStatus(node))
assert node.address == node_address == node_status["result"]["address"]

try:
    # tx = sdk.transactor.subscribe_to_gigabytes(node_address=node_address, gigabytes=1)
    tx = sdk.nodes.SubscribeToNode(
        node_address=node_address,
        gigabytes=1,
        hours=0,
        denom="udvpn",  # Comment this, raise exception
        tx_params=TxParams(),
    )
except grpc.RpcError as rpc_error:
    # TODO: this should be handled on skd side
    print(rpc_error.code())  # pylint: disable=no-member
    print(rpc_error.details())  # pylint: disable=no-member
    print(rpc_error.debug_error_string())  # pylint: disable=no-member

# The wait_for_tx can be founded in every modules because is inherited
# For example: tx_response = sdk.sessions.wait_for_tx(tx["hash"])
tx_response = sdk.nodes.wait_for_tx(tx["hash"])

subscription_id = search_attribute(
    tx_response, "sentinel.node.v2.EventCreateSubscription", "id"
)

# Double verify the subscription
subscription = sdk.subscriptions.QuerySubscription(
    int(subscription_id)
)
assert (
    subscription.base.id == int(subscription_id)
    and subscription.node_address == node_address
)

sessions = sdk.sessions.QuerySessionsForSubscription(
    subscription.base.id
)

# TODO: return message instead of executing the tx It's currently not supported ...

"""
messages = []
if sessions != []:
    for session in sessions:
        messages.append(
            sdk.transactor.PrepareOrTransactMsg(
                MsgEndRequest, id=session.id, rating=0, transact=False
            )
        )

messages.append(
    sdk.transactor.PrepareOrTransactMsg(
        MsgStartRequest, id=int(subscription_id), address=node_address, transact=False
    )
)

# TODO: Currently doesn't work, we should call self.__client.load_account_data, before each tx
try:
    tx = sdk.transactor.transaction(messages, tx_params=TxParams())
    tx_response = sdk.transactor.wait_for_tx(tx["hash"])
except grpc.RpcError as rpc_error:
    # TODO: this should be handled on skd side
    print(rpc_error.code())  # pylint: disable=no-member
    print(rpc_error.details())  # pylint: disable=no-member
    print(rpc_error.debug_error_string())  # pylint: disable=no-member
"""

for session in sessions:
    if session.status == Status.ACTIVE.value:
        tx = sdk.sessions.EndSession(session_id=session.id, rating=0)
        print(sdk.sessions.wait_for_tx(tx["hash"]))

tx = sdk.sessions.StartSession(subscription_id=int(subscription_id), address=node_address)
tx_response = sdk.sessions.wait_for_tx(tx["hash"])

session_id = search_attribute(
    tx_response, "sentinel.session.v2.EventStart", "id"
)
# Double verify the session
session = sdk.sessions.QuerySession(int(session_id))
assert (
    session.id == int(session_id)
    and session.node_address == node_address
    and session.subscription_id == int(subscription_id)
)

print(sdk.nodes.PostSession(int(session_id), node.remote_url, NodeType(node_status["result"]["type"])))



from typing import Any

import grpc
import sentinel_protobuf.sentinel.plan.v2.querier_pb2 as sentinel_plan_v2_querier_pb2
import sentinel_protobuf.sentinel.plan.v2.querier_pb2_grpc as sentinel_plan_v2_querier_pb2_grpc
import sentinel_protobuf.sentinel.plan.v2.msg_pb2 as msg_pb2

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.transactor.transactor import Transactor
from sentinel_sdk.types import PageRequest, TxParams


class PlanModule(Querier, Transactor):
    def __init__(self, channel: grpc.Channel, account, provider_account, client):
        self.__stub = sentinel_plan_v2_querier_pb2_grpc.QueryServiceStub(channel)
        self._account = account 
        self._provider_account = provider_account
        self._client = client

    def QueryPlan(self, plan_id: int) -> Any:
        try:
            r = self.__stub.QueryPlan(
                sentinel_plan_v2_querier_pb2.QueryPlanRequest(id=plan_id)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.plan

    def QueryPlansForProvider(
        self, address: str, status: int, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QueryPlansForProvider,
            request=sentinel_plan_v2_querier_pb2.QueryPlansForProviderRequest,
            attribute="plans",
            address=address,
            status=status.value,
            pagination=pagination,
        )

    def QueryPlans(self, status: int, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryPlans,
            request=sentinel_plan_v2_querier_pb2.QueryPlansRequest,
            attribute="plans",
            status=status.value,
            pagination=pagination,
        )

    def Create(self, duration: int, gigabytes: int, prices: int, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgCreateRequest(
            frm = self._account.address,
            duration = duration,
            gigabytes = gigabytes,
            prices = prices,
        )
        return self.transaction([msg], tx_params)

    def LinkNode(self, plan_id: int, node_address: str, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgLinkNodeRequest(
            frm = self._provider_account.address,
            id = plan_id,
            node_address = node_address,
        )
        return self.transaction([msg], tx_params)

    def Subscribe(self, denom: str, plan_id: int, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgSubscribeRequest(
            frm = self._account.address,
            id = plan_id,
            denom = denom,
        )
        return self.transaction([msg], tx_params)

    def UnlinkNode(self, plan_id: int, node_address: str, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgUnlinkNodeRequest(
            frm = self._provider_account.address,
            id = plan_id,
            node_address = node_address,
        )
        return self.transaction([msg], tx_params)

    def UpdateStatus(self, plan_id: int, status: int, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgUpdateStatusRequest(
            frm = self._provider_account.address,
            id = plan_id,
            status = status,
        )
        return self.transaction([msg], tx_params)



from typing import Any

import grpc
import sentinel_protobuf.sentinel.plan.v2.querier_pb2 as sentinel_plan_v2_querier_pb2
import sentinel_protobuf.sentinel.plan.v2.querier_pb2_grpc as sentinel_plan_v2_querier_pb2_grpc

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.types import PageRequest


class PlanQuerier(Querier):
    def __init__(self, channel: grpc.Channel):
        self.__stub = sentinel_plan_v2_querier_pb2_grpc.QueryServiceStub(channel)

    def QueryPlan(self, plan_id: int) -> Any:
        try:
            r = self.__stub.QueryPlan(
                sentinel_plan_v2_querier_pb2.QueryPlanRequest(id=plan_id)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.plan

    def QueryPlansForProvider(
        self, address: str, status: int, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QueryPlansForProvider,
            request=sentinel_plan_v2_querier_pb2.QueryPlansForProviderRequest,
            attribute="plans",
            address=address,
            status=status.value,
            pagination=pagination,
        )

    def QueryPlans(self, status: int, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryPlans,
            request=sentinel_plan_v2_querier_pb2.QueryPlansRequest,
            attribute="plans",
            status=status.value,
            pagination=pagination,
        )



import ssl
import threading
import urllib.parse
import urllib.request
import http.client
from typing import Any
import json
import uuid
import base64
import hashlib
import ecdsa
import grpc
import sentinel_protobuf.sentinel.node.v2.node_pb2 as node_pb2
import sentinel_protobuf.sentinel.node.v2.querier_pb2 as sentinel_node_v2_querier_pb2
import sentinel_protobuf.sentinel.node.v2.querier_pb2_grpc as sentinel_node_v2_querier_pb2_grpc
import sentinel_protobuf.sentinel.node.v2.msg_pb2 as msg_pb2

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.transactor.transactor import Transactor
from sentinel_sdk.types import PageRequest, TxParams, NodeType

from .wireguard import WgKey


class NodeModule(Querier, Transactor):
    def __init__(self, channel: grpc.Channel, node_timeout: int, account, client):
        self.node_timeout = node_timeout
        self.__stub = sentinel_node_v2_querier_pb2_grpc.QueryServiceStub(channel)

        # Disable SSL verification
        self.__ssl_ctx = ssl.create_default_context()
        self.__ssl_ctx.check_hostname = False
        self.__ssl_ctx.verify_mode = ssl.CERT_NONE

        self._account = account
        self._client = client

        self.__nodes_status_cache = {}

    def QueryParams(self) -> Any:
        return self.__stub.QueryParams(sentinel_node_v2_querier_pb2.QueryParamsRequest()).params

    def QueryNode(self, address: str) -> Any:
        r = self.__stub.QueryNode(
            sentinel_node_v2_querier_pb2.QueryNodeRequest(address=address)
        )
        return r.node

    def QueryNodes(self, status: int, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryNodes,
            request=sentinel_node_v2_querier_pb2.QueryNodesRequest,
            attribute="nodes",
            status=status.value,
            pagination=pagination,
        )

    def QueryNumOfNodesWithStatus(self, status: int) -> int:
        r = self.__stub.QueryNodes(
            sentinel_node_v2_querier_pb2.QueryNodesRequest(status=status.value)
        )
        return r.pagination.total

    def QueryNodeStatus(self, node: node_pb2.Node, is_in_thread: bool = False) -> str:
        node_endpoint = node.remote_url
        try:
            contents = urllib.request.urlopen(
                f"{node_endpoint}/status",
                context=self.__ssl_ctx,
                timeout=self.node_timeout,
            ).read()
            contents = contents.decode("utf-8")
        except urllib.error.URLError:
            contents = '{"success":false,"urllib-error":"URLError encountered"}'
        except TimeoutError:
            contents = '{"success":false,"urllib-error":"Data reading timed out"}'
        except http.client.RemoteDisconnected:
            contents = '{"success":false,"http-error":"Remote endpoint closed connection"}'
        except:
            contents = '{"success":false,"error":"Unrecognizable error encountered"}'
        if is_in_thread:
            self.__nodes_status_cache[node.address] = contents
        else:
            return contents

    def QueryNodesStatus(self, nodes: list[node_pb2.Node], n_threads: int = 8) -> dict:
        chunks = list(self.__split_into_chunks(nodes, n_threads))
        cur_threads = []
        for c in chunks:
            t = threading.Thread(target=self.__QueryNodesChunk, args=(c,))
            cur_threads.append(t)
            t.start()
        for t in cur_threads:
            t.join()

        result = self.__nodes_status_cache
        self.__nodes_status_cache = {}

        return result

    def QueryNodesForPlan(
        self, plan_id: int, status: int, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QueryNodesForPlan,
            request=sentinel_node_v2_querier_pb2.QueryNodesForPlanRequest,
            attribute="nodes",
            status=status.value,
            id=plan_id,
            pagination=pagination,
        )

    def __QueryNodesChunk(self, chunk: list[node_pb2.Node]):
        for node in chunk:
            self.QueryNodeStatus(node, True)

    def __split_into_chunks(self, items: list, size: int):
        for i in range(0, len(items), size):
            plus = i + size
            yield items[i:plus]
            # https://github.com/PyCQA/pycodestyle/issues/373

    def RegisterNode(self, gigabyte_prices: int, hourly_prices: int, remote_url: str, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgRegisterRequest(
            frm = self._account.address,
            gigabyte_prices=gigabyte_prices,
            hourly_prices=hourly_prices,
            remote_url=remote_url,
        )
        return self.transaction([msg], tx_params)

    def SubscribeToNode(self, node_address: str, gigabytes: int = 0, hours: int = 0, denom: str = "udvpn", tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgSubscribeRequest(
            frm = self._account.address,
            denom = denom,
            gigabytes = gigabytes,
            hours = hours,
            node_address = node_address,
        )
        return self.transaction([msg], tx_params)

    def UpdateNodeDetails(self, gigabyte_prices: int, hourly_prices: int, remote_url: str, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgUpdateDetailsRequest(
            frm = self._account.address,
            gigabyte_prices = gigabyte_prices,
            hourly_prices = hourly_prices,
            remote_url = remote_url,
        )
        return self.transaction([msg], tx_params)

    def UpdateNodeStatus(self, status: int, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgUpdateStatusRequest(
            frm = self._account.address,
            status = status.value,
        )
        return self.transaction([msg], tx_params)

    def __post_session(self, url: str, payload: dict) -> str:
        request = urllib.request.Request(url)
        request.add_header("Content-Type", "application/json; charset=utf-8")
        json_data_bytes = json.dumps(payload).encode("utf-8")  # needs to be bytes
        request.add_header("Content-Length", len(json_data_bytes))
        # TODO: status code != 200 was not handled
        with urllib.request.urlopen(request, json_data_bytes, timeout=self.node_timeout, context=self.__ssl_ctx) as f:
            return f.read().decode("utf-8")

    # TODO: find another 'fancy' name
    def PostSession(
        self, session_id: int, remote_url: str, node_type: NodeType = NodeType.WIREGUARD
    ):
        if node_type == NodeType.WIREGUARD:
            # [from golang] wgPrivateKey, err = wireguardtypes.NewPrivateKey()
            # [from golang] key = wgPrivateKey.Public().String()
            wgkey = WgKey()
            # The private key should be used by the wireguard client
            key = wgkey.pubkey
        else:  # NodeType.V2RAY
            # os.urandom(16)
            # [from golang] uid, err = uuid.GenerateRandomBytes(16)
            # [from golang] key = base64.StdEncoding.EncodeToString(append([]byte{0x01}, uid...))
            key = base64.b64encode(uuid.uuid4().bytes).decode("utf-8")

        # append([]byte{0x01}, uid) is required)
        # The key of wireguard is 32 bytes length, for v2ray only 16?

        # self._account inherited from Transactor class.
        # self._account.private_key and self._account.address

        sk = ecdsa.SigningKey.from_string(
            self._account.private_key, curve=ecdsa.SECP256k1, hashfunc=hashlib.sha256
        )

        session_id = int(session_id)
        # Uint64ToBigEndian
        bige_session = session_id.to_bytes(8, "big")
        signature = sk.sign(bige_session)

        payload = {
            "key": key,
            "signature": base64.b64encode(signature).decode("utf-8"),
        }
        return self.__post_session(
            f"{remote_url}/accounts/{self._account.address}/sessions/{session_id}",
            payload,
        )



import grpc

import sentinel_protobuf.cosmos.base.reflection.v1beta1.reflection_pb2 as cosmos_reflection_v1beta1_reflection_pb2
import sentinel_protobuf.cosmos.auth.v1beta1.auth_pb2 as cosmos_auth_v1beta1_auth_pb2
import sentinel_protobuf.cosmos.auth.v1beta1.query_pb2 as cosmos_auth_v1beta1_query_pb2

from sentinel_sdk.modules.deposit import DepositModule
from sentinel_sdk.modules.node import NodeModule
from sentinel_sdk.modules.plan import PlanModule
from sentinel_sdk.modules.swap import SwapModule
from sentinel_sdk.modules.provider import ProviderModule
from sentinel_sdk.modules.session import SessionModule
from sentinel_sdk.modules.subscription import SubscriptionModule

from mospy import Account
from mospy.clients import GRPCClient


from bip_utils import Bech32Encoder, Bip39SeedGenerator, Bip44, Bip44Coins, Bip39MnemonicValidator

from Crypto.Hash import RIPEMD160, SHA256

class SDKInstance:
    def __init__(self, grpcaddr: str, grpcport: int, secret: str = None, ssl: bool = False):
        try:
            self.__create_and_verify_channel(grpcaddr, grpcport, ssl=ssl)
        except grpc._channel._InactiveRpcError:
            raise ConnectionError("gRPC endpoint is invalid or not responding")

        self._client = None
        self._account = None
        self._provider_account = None
        
        if secret is not None:
            self.__setup_account_and_client(grpcaddr, grpcport, secret, ssl)

        self.__load_modules()


    def __create_and_verify_channel(
        self, grpcaddr: str, grpcport: int, ssl: bool = False
    ) -> grpc.Channel:
        channel = (
            grpc.secure_channel(
                f"{grpcaddr}:{grpcport}", credentials=grpc.ssl_channel_credentials()
            )
            if ssl is True
            else grpc.insecure_channel(f"{grpcaddr}:{grpcport}")
        )

        tmp_stub = channel.unary_unary(
            "/cosmos.base.reflection.v1beta1.ReflectionService/ListAllInterfaces",
            request_serializer=cosmos_reflection_v1beta1_reflection_pb2.ListAllInterfacesRequest.SerializeToString,
            response_deserializer=cosmos_reflection_v1beta1_reflection_pb2.ListAllInterfacesResponse.FromString,
        )

        # Test if the grpc is valid and respond to ListAllInterfacesRequest
        tmp_stub(cosmos_reflection_v1beta1_reflection_pb2.ListAllInterfacesRequest())
        self._channel = channel

    def __setup_account_and_client(self, grpcaddr: str, grpcport: int, secret: str, use_ssl: bool = False):
        self._account = self.__create_account(secret)
        self._provider_account = self.__create_provider_account(secret)
        self._client = self.__create_client(grpcaddr, grpcport, use_ssl)
        self._client.load_account_data(account=self._account)

    def __create_account(self, secret: str):
        try:
            Bip39MnemonicValidator().Validate(secret)
            seed_bytes = Bip39SeedGenerator(secret).Generate()
            bip44_def_ctx = Bip44.FromSeed(seed_bytes, Bip44Coins.COSMOS).DeriveDefaultPath()
        except:
            try:
                int(secret, 16)
                bip44_def_ctx = Bip44.FromPrivateKey(bytes.fromhex(secret), Bip44Coins.COSMOS)
            except:
                raise ValueError("Unrecognized secret either as a mnemonic or hex private key")

        sha_key = SHA256.new()
        ripemd_key = RIPEMD160.new()
        sha_key.update(bip44_def_ctx.PublicKey().RawCompressed().m_data_bytes)
        ripemd_key.update(sha_key.digest())
        bech32_pub = Bech32Encoder.Encode("sent", ripemd_key.digest())
        account_num = self.__get_account_number(bech32_pub)
        account = Account(
            private_key=bip44_def_ctx.PrivateKey().Raw().ToHex(),
            hrp="sent",
            account_number=account_num,
            protobuf="sentinel",
        )
        return account
    
    def __create_provider_account(self, secret: str):
        try:
            Bip39MnemonicValidator().Validate(secret)
            seed_bytes = Bip39SeedGenerator(secret).Generate()
            bip44_def_ctx = Bip44.FromSeed(seed_bytes, Bip44Coins.COSMOS).DeriveDefaultPath()
        except:
            try:
                int(secret, 16)
                bip44_def_ctx = Bip44.FromPrivateKey(bytes.fromhex(secret), Bip44Coins.COSMOS)
            except:
                raise ValueError("Unrecognized secret either as a mnemonic or hex private key")

        sha_key = SHA256.new()
        ripemd_key = RIPEMD160.new()
        sha_key.update(bip44_def_ctx.PublicKey().RawCompressed().m_data_bytes)
        ripemd_key.update(sha_key.digest())
        bech32_pub = Bech32Encoder.Encode("sent", ripemd_key.digest())
        account_num = self.__get_account_number(bech32_pub)
        account = Account(
            private_key=bip44_def_ctx.PrivateKey().Raw().ToHex(),
            hrp="sentprov",
            account_number=account_num,
            protobuf="sentinel",
        )
        return account

    def __create_client(self, grpcaddr: str, grpcport: int, use_ssl: bool = False):
        client = GRPCClient(
            host=grpcaddr, port=grpcport, ssl=use_ssl, protobuf="sentinel"
        )
        return client

    def __get_account_number(self, address: str):
        tmp_stub = self._channel.unary_unary(
            "/cosmos.auth.v1beta1.Query/Account",
            request_serializer=cosmos_auth_v1beta1_query_pb2.QueryAccountRequest.SerializeToString,
            response_deserializer=cosmos_auth_v1beta1_query_pb2.QueryAccountResponse.FromString,
        )
        r = tmp_stub(cosmos_auth_v1beta1_query_pb2.QueryAccountRequest(address=address))
        baseacc = cosmos_auth_v1beta1_auth_pb2.BaseAccount()
        baseacc.ParseFromString(r.account.value)
        return baseacc.account_number


    def __load_modules(self):
        self.nodes = NodeModule(self._channel, 10, self._account, self._client)
        self.deposits = DepositModule(self._channel)
        self.plans = PlanModule(self._channel, self._account, self._provider_account, self._client)
        self.providers = ProviderModule(self._channel, self._account, self._client)
        self.sessions = SessionModule(self._channel, self._account, self._client)
        self.subscriptions = SubscriptionModule(self._channel, self._account, self._client)
        self.swaps = SwapModule(self._channel, self._account, self._client)

    def renew_account(self, secret: str):
        self._account = self.__create_account(secret)
        self._client.load_account_data(account=self._account)
        self.__load_modules()

    def renew_grpc(self, grpcaddr: str, grpcport: int, use_ssl: bool = False):
        self._client = self.__create_client(grpcaddr, grpcport)
        self._client.load_account_data(account=self._account)
        self.__load_modules()
        
    @property
    def grpcaddr(self):
        return self._client._host
        
    @property
    def grpcport(self):
        return self._client._port

        



import random

from sentinel_sdk.sdk import SDKInstance
from sentinel_sdk.types import PageRequest, Status

# GRPC_VERBOSITY=debug GRPC_TRACE=tcp,http python test/queryall.py

sdk = SDKInstance("grpc.sentinel.co", 9090)

nodes = sdk.multiquerier.node_querier.QueryNodes(
    Status.ACTIVE, pagination=PageRequest(limit=500)
)
print(f"{len(nodes)} nodes")
node_random = random.choice(nodes)
node_chain = sdk.multiquerier.node_querier.QueryNode(node_random.address)
assert node_chain == node_random
print(f"Random node: {node_chain}")

plans = sdk.multiquerier.plan_querier.QueryPlans(Status.ACTIVE)
print(f"{len(plans)} plans")
plan_random = random.choice(plans)
plan_chain = sdk.multiquerier.plan_querier.QueryPlan(plan_random.id)
assert plan_chain == plan_random
print(f"Random plan: {plan_chain}")

nodes_4plan = sdk.multiquerier.node_querier.QueryNodesForPlan(
    plan_random.id, Status.ACTIVE
)
print(f"{len(nodes_4plan)} nodes for plan: {plan_random.id}")

deposits = sdk.multiquerier.deposit_querier.QueryDeposits()
print(f"{len(deposits)} deposits")
deposit_random = random.choice(deposits)
deposit_chain = sdk.multiquerier.deposit_querier.QueryDeposit(deposit_random.address)
assert deposit_chain == deposit_random
print(f"Random deposit: {deposit_chain}")

providers = sdk.multiquerier.provider_querier.QueryProviders(Status.ACTIVE)
print(f"{len(providers)} providers")
provider_random = random.choice(providers)
provider_chain = sdk.multiquerier.provider_querier.QueryProvider(
    provider_random.address
)
assert provider_chain == provider_random
print(f"Random provider: {provider_chain}")

subscriptions = sdk.multiquerier.subscription_querier.QuerySubscriptions(
    pagination=PageRequest(limit=1500)
)
print(f"{len(subscriptions)} subscriptions")
subscription_random = random.choice(subscriptions)
subscription_chain = sdk.multiquerier.subscription_querier.QuerySubscription(
    subscription_random.base.id
)
assert subscription_chain == subscription_random
print(f"Random subscription: {subscription_chain}")

# Not all subscriptions have an allocations, going to search randomly
allocations = []
while allocations == []:
    allocations = sdk.multiquerier.subscription_querier.QueryAllocations(
        random.choice(subscriptions).base.id
    )
    print(f"{len(allocations)} allocations")

allocation_random = random.choice(allocations)
allocation_chain = sdk.multiquerier.subscription_querier.QueryAllocation(
    allocation_random.address, allocation_random.id
)
assert allocation_chain == allocation_random
print(f"Random allocation: {allocation_chain}")

sessions = sdk.multiquerier.session_querier.QuerySessions()
print(f"{len(sessions)} sessions")
session_random = random.choice(sessions)
session_chain = sdk.multiquerier.session_querier.QuerySession(session_random.id)
assert session_chain == session_random
print(f"Random session: {session_chain}")

payouts = sdk.multiquerier.subscription_querier.QueryPayouts()
print(f"{len(payouts)} payouts")
payout_random = random.choice(payouts)
payout_chain = sdk.multiquerier.subscription_querier.QueryPayout(payout_random.id)
assert payout_chain == payout_random
print(f"Random payout: {payout_chain}")

# TODO: not sure about this query
payouts = sdk.multiquerier.subscription_querier.QueryPayoutsForAccount(
    subscription_random.base.address
)
print(
    f"{len(payouts)} payouts, QueryPayoutsForAccount({subscription_random.base.address})"
)

payouts = sdk.multiquerier.subscription_querier.QueryPayoutsForNode(
    subscription_random.node_address
)
print(
    f"{len(payouts)} payouts, QueryPayoutsForNode({subscription_random.node_address})"
)

subscriptions = sdk.multiquerier.subscription_querier.QuerySubscriptionsForAccount(
    subscription_random.base.address
)
print(
    f"{len(subscriptions)} subscriptions, QuerySubscriptionsForAccount({subscription_random.base.address})"
)

subscriptions = sdk.multiquerier.subscription_querier.QuerySubscriptionsForNode(
    subscription_random.node_address
)
print(
    f"{len(subscriptions)} subscriptions, QuerySubscriptionsForNode({subscription_random.node_address})"
)

subscriptions = sdk.multiquerier.subscription_querier.QuerySubscriptionsForPlan(
    plan_random.id
)
print(
    f"{len(subscriptions)} subscriptions, QuerySubscriptionsForPlan({plan_random.id})"
)


sessions = sdk.multiquerier.session_querier.QuerySessionsForAccount(
    subscription_random.base.address
)
print(
    f"{len(sessions)} sessions, QuerySessionsForAccount({subscription_random.base.address})"
)

sessions = sdk.multiquerier.session_querier.QuerySessionsForAllocation(
    allocation_random.id, allocation_random.address
)
print(
    f"{len(sessions)} sessions, QuerySessionsForAllocation({allocation_random.id}, {allocation_random.address})"
)

sessions = sdk.multiquerier.session_querier.QuerySessionsForNode(
    subscription_random.node_address
)
print(
    f"{len(sessions)} sessions, QuerySessionsForNode({subscription_random.node_address})"
)

sessions = sdk.multiquerier.session_querier.QuerySessionsForSubscription(
    subscription_random.base.id
)
print(
    f"{len(sessions)} sessions, QuerySessionsForSubscription({subscription_random.base.id})"
)



from typing import Any

import grpc
import sentinel_protobuf.sentinel.swap.v1.querier_pb2 as sentinel_swap_v1_querier_pb2
import sentinel_protobuf.sentinel.swap.v1.querier_pb2_grpc as sentinel_swap_v1_querier_pb2_grpc

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.types import PageRequest

# import sentinel_protobuf.sentinel.swap.v1.swap_pb2 as swap_pb2


class SwapQuerier(Querier):
    def __init__(self, channel: grpc.Channel):
        self.__stub = sentinel_swap_v1_querier_pb2_grpc.QueryServiceStub(channel)

    def QuerySwap(self, tx_hash: bytes) -> Any:
        try:
            r = self.__stub.QuerySwap(
                sentinel_swap_v1_querier_pb2.QuerySwapRequest(tx_hash=tx_hash)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.swap

    def QuerySwaps(self, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QuerySwaps,
            request=sentinel_swap_v1_querier_pb2.QuerySwapsRequest,
            attribute="swaps",
            pagination=pagination,
        )



import grpc

from sentinel_sdk.querier.modules.deposit import DepositQuerier
from sentinel_sdk.querier.modules.node import NodeQuerier
from sentinel_sdk.querier.modules.plan import PlanQuerier
from sentinel_sdk.querier.modules.provider import ProviderQuerier
from sentinel_sdk.querier.modules.session import SessionQuerier
from sentinel_sdk.querier.modules.subscription import SubscriptionQuerier
from sentinel_sdk.querier.modules.swap import SwapQuerier


class SentinelQuerier:
    def __init__(self, channel: grpc.Channel):
        self._channel = channel
        self.node_querier = NodeQuerier(self._channel, 10)
        self.deposit_querier = DepositQuerier(self._channel)
        self.plan_querier = PlanQuerier(self._channel)
        self.provider_querier = ProviderQuerier(self._channel)
        self.session_querier = SessionQuerier(self._channel)
        self.subscription_querier = SubscriptionQuerier(self._channel)
        self.swap_querier = SwapQuerier(self._channel)



from typing import Any

from sentinel_sdk.types import PageRequest


class Querier:
    def QueryAll(
        self,
        query: Any,
        request: Any,
        attribute: str,
        pagination: PageRequest = None,
        **kwargs,
    ) -> list:
        # ["status", "address", "id"] are the only valid kwargs
        request_arguments = {
            k: kwargs[k] for k in kwargs if k in ["status", "address", "id"]
        }

        if pagination is None:
            # Create an empty page request, will be used for pagination with next_key
            # On first iteration with next_key = 0x01, the pagination will have only limit value
            pagination = PageRequest()

        items = []
        next_key = 0x01

        while next_key:
            request_arguments["pagination"] = pagination.build(key=next_key)
            r = query(request(**request_arguments))
            next_key = r.pagination.next_key
            items += [item for item in getattr(r, attribute)]

            # Stop iteration if PageRequest contain a offset value
            if pagination.offset is not None:
                break

        return items


import ssl
import threading
import urllib.request
from typing import Any

import grpc
import sentinel_protobuf.sentinel.node.v2.node_pb2 as node_pb2
import sentinel_protobuf.sentinel.node.v2.querier_pb2 as sentinel_node_v2_querier_pb2
import sentinel_protobuf.sentinel.node.v2.querier_pb2_grpc as sentinel_node_v2_querier_pb2_grpc

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.types import PageRequest


class NodeQuerier(Querier):
    def __init__(self, channel: grpc.Channel, status_fetch_timeout: int):
        self.status_fetch_timeout = status_fetch_timeout
        self.__stub = sentinel_node_v2_querier_pb2_grpc.QueryServiceStub(channel)

        # Disable SSL verification
        self.__ssl_ctx = ssl.create_default_context()
        self.__ssl_ctx.check_hostname = False
        self.__ssl_ctx.verify_mode = ssl.CERT_NONE

        self.__nodes_status_cache = {}

    def QueryNode(self, address: str) -> Any:
        r = self.__stub.QueryNode(
            sentinel_node_v2_querier_pb2.QueryNodeRequest(address=address)
        )
        return r.node

    def QueryNodes(self, status: int, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryNodes,
            request=sentinel_node_v2_querier_pb2.QueryNodesRequest,
            attribute="nodes",
            status=status.value,
            pagination=pagination,
        )

    def QueryNumOfNodesWithStatus(self, status: int) -> int:
        r = self.__stub.QueryNodes(
            sentinel_node_v2_querier_pb2.QueryNodesRequest(status=status.value)
        )
        return r.pagination.total

    def QueryNodeStatus(self, node: node_pb2.Node, is_in_thread: bool = False) -> str:
        node_endpoint = node.remote_url
        try:
            contents = urllib.request.urlopen(
                f"{node_endpoint}/status",
                context=self.__ssl_ctx,
                timeout=self.status_fetch_timeout,
            ).read()
            contents = contents.decode("utf-8")
        except urllib.error.URLError:
            contents = '{"success":false,"urllib-error":"URLError encountered"}'
        except TimeoutError:
            contents = '{"success":false,"urllib-error":"Data reading timed out"}'

        if is_in_thread:
            self.__nodes_status_cache[node.address] = contents
        else:
            return contents

    def QueryNodesStatus(self, nodes: list[node_pb2.Node], n_threads: int = 8) -> dict:
        chunks = list(self.__split_into_chunks(nodes, n_threads))
        cur_threads = []
        for c in chunks:
            t = threading.Thread(target=self.__QueryNodesChunk, args=(c,))
            cur_threads.append(t)
            t.start()
        for t in cur_threads:
            t.join()

        result = self.__nodes_status_cache
        self.__nodes_status_cache = {}

        return result

    def QueryNodesForPlan(
        self, plan_id: int, status: int, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QueryNodesForPlan,
            request=sentinel_node_v2_querier_pb2.QueryNodesForPlanRequest,
            attribute="nodes",
            status=status.value,
            id=plan_id,
            pagination=pagination,
        )

    def __QueryNodesChunk(self, chunk: list[node_pb2.Node]):
        for node in chunk:
            self.QueryNodeStatus(node, True)

    def __split_into_chunks(self, items: list, size: int):
        for i in range(0, len(items), size):
            plus = i + size
            yield items[i:plus]
            # https://github.com/PyCQA/pycodestyle/issues/373



from typing import Any

import grpc
import sentinel_protobuf.sentinel.provider.v2.querier_pb2 as sentinel_provider_v2_querier_pb2
import sentinel_protobuf.sentinel.provider.v2.querier_pb2_grpc as sentinel_provider_v2_querier_pb2_grpc

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.types import PageRequest


class ProviderQuerier(Querier):
    def __init__(self, channel: grpc.Channel):
        self.__stub = sentinel_provider_v2_querier_pb2_grpc.QueryServiceStub(channel)

    def QueryProvider(self, address: str) -> Any:
        try:
            r = self.__stub.QueryProvider(
                sentinel_provider_v2_querier_pb2.QueryProviderRequest(address=address)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.provider

    def QueryProviders(self, status: int, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryProviders,
            request=sentinel_provider_v2_querier_pb2.QueryProvidersRequest,
            attribute="providers",
            status=status.value,
            pagination=pagination,
        )



import json
import base64

# TODO: we want to move this method (?)
# In the latest version of mospy the response was converted as dict
# from google.protobuf.json_format import MessageToDict
# https://github.com/ctrl-Felix/mospy/blob/ea07705b6ec4c76cc355b7dc933fae7c09fd8429/src/mospy/clients/GRPCClient.py#L155
def search_attribute(tx_response: dict, event_type: str, attribute_key: str) -> dict:
    for event in tx_response.get("txResponse", tx_response).get("events", []):
        if event["type"] == event_type:
            for attribute in event["attributes"]:
                if base64.b64decode(attribute["key"]) == attribute_key.encode():
                    return json.loads(base64.b64decode(attribute["value"]).decode("utf-8"))
    return None



# Tkd-Alex
# Python implementation of: https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go

"""
# Too much issue with cosmpy and sentinel_protobuf / goproto import duplicated
from cosmpy.aerial.tx import Transaction
from cosmpy.aerial.tx_helpers import TxResponse
from cosmpy.aerial.client.utils import prepare_and_broadcast_basic_transaction
from cosmpy.aerial.client import LedgerClient, NetworkConfig

cfg = NetworkConfig(
    chain_id="sentinelhub-2",
    url="grpc+http://aimokoivunen.mathnodes.com:9090/",
    fee_minimum_gas_price=0.4,
    fee_denomination="udvpn",
    staking_denomination="udvpn",
)
client = LedgerClient(cfg)
"""

import base64
import hashlib
import json
import ssl
import urllib.parse
import urllib.request

import ecdsa
import grpc
from bip_utils import Bip39SeedGenerator, Bip44, Bip44Coins

# pip install git+https://github.com/Tkd-Alex/mospy.git
from mospy import Account, Transaction
from mospy.clients import GRPCClient
from pywgkey import WgKey
from sentinel_protobuf.cosmos.base.v1beta1.coin_pb2 import Coin
from sentinel_protobuf.sentinel.node.v2.msg_pb2 import MsgSubscribeRequest
from sentinel_protobuf.sentinel.node.v2.querier_pb2 import QueryNodeRequest
from sentinel_protobuf.sentinel.node.v2.querier_pb2_grpc import (
    QueryServiceStub as QueryNodeServiceStub,
)
from sentinel_protobuf.sentinel.session.v2.msg_pb2 import MsgStartRequest
from sentinel_protobuf.sentinel.session.v2.querier_pb2 import (
    QuerySessionsForAllocationRequest,
)
from sentinel_protobuf.sentinel.session.v2.querier_pb2_grpc import (
    QueryServiceStub as QuerySessionServiceStub,
)

# from mospy.clients import HTTPClient

# Custom version of mospy
# Tecnically we could re-implement mospy on this repository (https://github.com/MathNodes/sentinel-python-sdk)


def fetch_node_info(url: str) -> dict:
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    with urllib.request.urlopen(f"{url}/status", timeout=(60 * 60), context=ctx) as f:
        return json.loads(f.read().decode("utf-8"))


def post_session_node(url: str, payload: dict) -> dict:
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    request = urllib.request.Request(url)
    request.add_header("Content-Type", "application/json; charset=utf-8")
    json_data_bytes = json.dumps(payload).encode("utf-8")  # needs to be bytes
    request.add_header("Content-Length", len(json_data_bytes))
    with urllib.request.urlopen(
        request, json_data_bytes, timeout=(60 * 60), context=ctx
    ) as f:
        return json.loads(f.read().decode("utf-8"))


grpc_host = "grpc.sentinel.co"
grpc_port = 9090

node_address = "sentnode123456789abcdefghilmopqrts"

mnemonic = "dont share your secret"
seed_bytes = Bip39SeedGenerator(mnemonic).Generate()
bip44_def_ctx = Bip44.FromSeed(seed_bytes, Bip44Coins.COSMOS).DeriveDefaultPath()

account = Account(
    private_key=bip44_def_ctx.PrivateKey().Raw().ToHex(),
    hrp="sent",
    account_number="38997",
    protobuf="sentinel",
)

# Subscribe to node request
# client = HTTPClient(api="https://api.sentinel.mathnodes.com")
client = GRPCClient(host=grpc_host, port=grpc_port, ssl=False, protobuf="sentinel")
client.load_account_data(account=account)

fee = Coin(denom="udvpn", amount="20000")
tx = Transaction(
    account=account, fee=fee, gas=0, protobuf="sentinel", chain_id="sentinelhub-2"
)
wmsg = MsgSubscribeRequest(
    frm=account.address, node_address=node_address, gigabytes=1, hours=0, denom="udvpn"
)
tx.add_raw_msg(wmsg, type_url="/sentinel.node.v2.MsgSubscribeRequest")
client.estimate_gas(transaction=tx, update=True, multiplier=1.5)
tx_response = client.broadcast_transaction(transaction=tx)

"""
con = client._connect()
tx_stub = client._cosmos_tx_service_pb2_grpc.ServiceStub(con)
tx_response = tx_stub.GetTx(client._cosmos_tx_service_pb2.GetTxRequest(hash=tx_response["hash"]))
"""

"""
from sentinel_protobuf.cosmos.tx.v1beta1.service_pb2_grpc import ServiceStub
from sentinel_protobuf.cosmos.tx.v1beta1.service_pb2 import GetTxRequest
channel = grpc.insecure_channel(f"{grpc_host}:{grpc_port}")
stub = ServiceStub(con)
tx_response = stub.GetTx(GetTxRequest(hash=tx_response["hash"]))
"""

# Maybe wait the tx before query ...
tx_response = client.get_tx(tx_response["hash"])

# Looking for subscription_id (sentinel.node.v2.EventCreateSubscription)
# We can do better, for the moment just iterate over the events
subscription_id = None
for event in tx_response.tx_response.events:
    if event.type == "sentinel.node.v2.EventCreateSubscription":
        for attribute in event.attributes:
            if attribute.key == b"id":
                # Why the hell the values are encoded grrr
                subscription_id = json.loads(attribute.value)
        break


channel = grpc.insecure_channel(f"{grpc_host}:{grpc_port}")

# Query node
# https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go#L192-L197
# Please use the sdk here....


stub = QueryNodeServiceStub(channel)

response = stub.QueryNode(QueryNodeRequest(address=node_address))
remote_url = response.node.remote_url

# Fetch info, in order to determine node type V2Ray or Wireguard
# Currently on wireguard is implement in this script
# Btw, for V2Ray the key It's just a random uuid
# https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go#L199-L202
node_info = fetch_node_info(remote_url)

# Query active session >> Sentinecli still user v1 query :/
# https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go#L210-L213
# from sentinel_protobuf.sentinel.session.v1.querier_pb2 import QuerySessionsForAddressRequest
# from sentinel_protobuf.sentinel.session.v1.querier_pb2_grpc import QueryNodeServiceStub
# No querirer for v1 :(
# Please use the sdk here....


stub = QuerySessionServiceStub(channel)
response = stub.QuerySessionsForAllocation(
    QuerySessionsForAllocationRequest(id=int(subscription_id), address=account.address)
)

# If session already exist send NewMsgEndRequest
# Bypass for the moment ....
# https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go#L215-L224

# Start new session NewMsgStartRequest
# https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go#L225-L237

client = GRPCClient(host=grpc_host, port=grpc_port, ssl=False, protobuf="sentinel")
# load_account_data required for each new tx
client.load_account_data(account=account)

fee = Coin(denom="udvpn", amount="20000")
tx = Transaction(
    account=account, fee=fee, gas=0, protobuf="sentinel", chain_id="sentinelhub-2"
)
wmsg = MsgStartRequest(
    frm=account.address, id=int(subscription_id), address=node_address
)
tx.add_raw_msg(wmsg, type_url="/sentinel.session.v2.MsgStartRequest")
client.estimate_gas(transaction=tx, update=True, multiplier=1.5)
tx_response = client.broadcast_transaction(transaction=tx)
# Maybe wait the tx before query ...
tx_response = client.get_tx(tx_response["hash"])

# Looking for session_id (sentinel.node.v2.EventCreateSubscription)
# We can do better, for the moment just iterate over the events
session_id = None
for event in tx_response.tx_response.events:
    if event.type == "sentinel.session.v2.EventStart":
        for attribute in event.attributes:
            if attribute.key == b"id":
                # Why the hell the values are encoded grrr
                session_id = json.loads(attribute.value)
        break

# Query again active session, we should have the same of session_id
# https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go#L239-L245
# Please use the sdk here....

stub = QuerySessionServiceStub(channel)
response = stub.QuerySessionsForAllocation(
    QuerySessionsForAllocationRequest(id=int(subscription_id), address=account.address)
)

for session in response.sessions:
    if session.subscription_id == int(subscription_id):
        if session.id == int(session_id):
            print("Session exist", session)
            break

# https://github.com/sentinel-official/cli-client/blob/master/cmd/connect.go#L253-L306
wgkey = WgKey()
private = wgkey.privkey
public = wgkey.pubkey

prv_key = bip44_def_ctx.PrivateKey().Raw().ToBytes()
sk = ecdsa.SigningKey.from_string(
    prv_key, curve=ecdsa.SECP256k1, hashfunc=hashlib.sha256
)

session_id = int(session_id)
# Uint64ToBigEndian
bige_session = session_id.to_bytes(8, "big")
signature = sk.sign(bige_session)

payload = {"key": f"{public}", "signature": base64.b64encode(signature).decode("utf-8")}

response = post_session_node(
    f"{remote_url}/accounts/{account.address}/sessions/{session_id}", payload
)
print(response)



import grpc
import ssl
import time
from typing import Any
from sentinel_protobuf.cosmos.base.v1beta1.coin_pb2 import Coin
from sentinel_sdk.types import NodeType, TxParams
from mospy import Transaction

class Transactor:
    def __init__(
        self,
        account,
        client
    ):
        self._account = account
        self._client = client

    # TODO: subscribe_to_gigabytes, subscribe_to_hours, start_request and all method should have fee denom and amount as args?
    # We need to find a good way to this. **The method could return only the MsgRequest and type_url**
    # Who wants to submit the tx, can call transaction (we should also implement multi-add_raw_msg), one tx with multiple msg

    def PrepareOrTransactMsg(   self,
                                msg: Any,
                                transact: bool = False,
                                tx_params : TxParams = TxParams(),
                                **kwargs):
        msg_args = {
            k: kwargs[k] for k in kwargs if k in ["address", "node_address", "id", "gigabytes", "hours", "rating", "denom"]
        }

        msg_args['frm'] = self.__account.address
        prepared_msg = msg(**msg_args)

        return self.transaction([prepared_msg], tx_params) if transact else prepared_msg


    def transaction(
        self,
        messages: list,
        tx_params: TxParams = TxParams(),
    ) -> dict:
        if self._account is None or self._client is None:
            raise ValueError("Transactor was not initialized due missing secret, unable to transact")

        tx = Transaction(
            account=self._account,
            fee=Coin(denom=tx_params.denom, amount=f"{tx_params.fee_amount}"),
            gas=tx_params.gas,
            protobuf="sentinel",
            chain_id="sentinelhub-2",
        )

        if not isinstance(messages, list):
            messages = [messages]

        for message in messages:
            tx.add_raw_msg(message, type_url=f"/{message.DESCRIPTOR.full_name}")

        # Required before each tx of we get account sequence mismatch, expected 945, got 944: incorrect account sequence
        self._client.load_account_data(account=self._account)

        # inplace, auto-update gas with update=True
        # auto calculate the gas only if was not already passed as args:
        if tx_params.gas == 0:
            self._client.estimate_gas(
                transaction=tx, update=True, multiplier=tx_params.gas_multiplier
            )

        tx_response = self._client.broadcast_transaction(transaction=tx)
        # tx_response = {"hash": hash, "code": code, "log": log}
        return tx_response

    def wait_for_tx(self, tx_hash: str, timeout: float = 120, poll_period: float = 10):
        if self._account is None or self._client is None:
            raise ValueError("Transactor was not initialized due missing secret, unable to wait transaction")

        return self._client.wait_for_tx(
            tx_hash=tx_hash,
            timeout=timeout,
            poll_period=poll_period
        )

    # back-compatibility
    def wait_transaction(self, tx_hash: str, timeout: float = 120, poll_period: float = 10):
        return self.wait_for_tx(tx_hash=tx_hash, timeout=timeout, poll_period=poll_period)



from base64 import b64encode
from nacl.public import PrivateKey
 
class WgKey():
    
    def __init__(self):
        self._key = PrivateKey.generate()
        
    @property
    def pubkey(self) -> str:
        return b64encode(bytes(self._key.public_key)).decode("ascii")
    
    @property
    def privkey(self) -> str:
        return b64encode(bytes(self._key)).decode("ascii")


from typing import Any

import grpc
import sentinel_protobuf.sentinel.session.v2.querier_pb2 as sentinel_session_v2_querier_pb2
import sentinel_protobuf.sentinel.session.v2.querier_pb2_grpc as sentinel_session_v2_querier_pb2_grpc

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.types import PageRequest


class SessionQuerier(Querier):
    def __init__(self, channel: grpc.Channel):
        self.__stub = sentinel_session_v2_querier_pb2_grpc.QueryServiceStub(channel)

    def QuerySession(self, session_id: int) -> Any:
        try:
            r = self.__stub.QuerySession(
                sentinel_session_v2_querier_pb2.QuerySessionRequest(id=session_id)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.session

    def QuerySessions(self, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QuerySessions,
            request=sentinel_session_v2_querier_pb2.QuerySessionsRequest,
            attribute="sessions",
            pagination=pagination,
        )

    def QuerySessionsForAccount(
        self, address: str, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QuerySessionsForAccount,
            request=sentinel_session_v2_querier_pb2.QuerySessionsForAccountRequest,
            attribute="sessions",
            address=address,
            pagination=pagination,
        )

    def QuerySessionsForAllocation(
        self, allocation_id: int, address: str, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QuerySessionsForAllocation,
            request=sentinel_session_v2_querier_pb2.QuerySessionsForAllocationRequest,
            attribute="sessions",
            id=allocation_id,
            address=address,
            pagination=pagination,
        )

    def QuerySessionsForNode(
        self, address: str, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QuerySessionsForNode,
            request=sentinel_session_v2_querier_pb2.QuerySessionsForNodeRequest,
            attribute="sessions",
            address=address,
            pagination=pagination,
        )

    def QuerySessionsForSubscription(
        self, subscription_id: int, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QuerySessionsForSubscription,
            request=sentinel_session_v2_querier_pb2.QuerySessionsForSubscriptionRequest,
            attribute="sessions",
            id=subscription_id,
            pagination=pagination,
        )



from typing import Any

import grpc
import sentinel_protobuf.sentinel.provider.v2.querier_pb2 as sentinel_provider_v2_querier_pb2
import sentinel_protobuf.sentinel.provider.v2.querier_pb2_grpc as sentinel_provider_v2_querier_pb2_grpc
import sentinel_protobuf.sentinel.provider.v2.msg_pb2 as msg_pb2

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.transactor.transactor import Transactor
from sentinel_sdk.types import PageRequest, TxParams


class ProviderModule(Querier, Transactor):
    def __init__(self, channel: grpc.Channel, account, client):
        self.__stub = sentinel_provider_v2_querier_pb2_grpc.QueryServiceStub(channel)
        self._account = account
        self._client = client

    def QueryProvider(self, address: str) -> Any:
        try:
            r = self.__stub.QueryProvider(
                sentinel_provider_v2_querier_pb2.QueryProviderRequest(address=address)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.provider

    def QueryProviders(self, status: int, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryProviders,
            request=sentinel_provider_v2_querier_pb2.QueryProvidersRequest,
            attribute="providers",
            status=status.value,
            pagination=pagination,
        )

    def Register(self, description: str, identity: str, name: str, website: str, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgRegisterRequest(
            frm = self._account.address,
            description = description,
            identity = identity,
            name = name,
            website = website,
        )
        return self.transaction([msg], tx_params)

    def Update(self, description: str, identity: str, name: str, website: str, tx_params: TxParams = TxParams()):
        msg = msg_pb2.MsgUpdateRequest(
            frm = self._account.address,
            description = description,
            identity = identity,
            name = name,
            website = website,
        )
        return self.transaction([msg], tx_params)


# https://github.com/MathNodes/cosmospy-protobuf/blob/chain/sentinel/src/sentinel_protobuf/sentinel/types/v1/status.proto
from enum import Enum
# https://github.com/cosmos/cosmos-sdk/blob/main/types/query/pagination.pb.go
import sentinel_protobuf.cosmos.base.query.v1beta1.pagination_pb2 as cosmos_pagination_pb2

from dataclasses import dataclass

class Status(Enum):
    UNSPECIFIED = 0
    ACTIVE = 1
    INACTIVE_PENDING = 2
    INACTIVE = 3

    def __str__(self):
        return f"Status{self.name.title()}"


class NodeType(Enum):
    WIREGUARD = 1
    V2RAY = 2

@dataclass
class TxParams:
    denom: str = "udvpn"
    fee_amount: int = 314159
    gas: float = 0
    gas_multiplier: float = 1.5


class PageRequest:
    """
    // key is a value returned in PageResponse.next_key to begin
    // querying the next page most efficiently. Only one of offset or key should be set.
    Key []byte `protobuf:"bytes,1,opt,name=key,proto3" json:"key,omitempty"`

    // offset is a numeric offset that can be used when key is unavailable.
    // It is less efficient than using key. Only one of offset or key should be set.
    Offset uint64 `protobuf:"varint,2,opt,name=offset,proto3" json:"offset,omitempty"`

    // limit is the total number of results to be returned in the result page.
    // If left empty it will default to a value to be set by each app.
    Limit uint64 `protobuf:"varint,3,opt,name=limit,proto3" json:"limit,omitempty"`

    // count_total is set to true  to indicate that the result set should include
    // a count of the total number of items available for pagination in UIs.
    // count_total is only respected when offset is used. It is ignored when key is set.
    CountTotal bool `protobuf:"varint,4,opt,name=count_total,json=countTotal,proto3" json:"count_total,omitempty"`

    // reverse is set to true if results are to be returned in the descending order.
    // Since: cosmos-sdk 0.43
    Reverse bool `protobuf:"varint,5,opt,name=reverse,proto3" json:"reverse,omitempty"`

    # https://github.com/sentinel-official/hub/blob/development/x/node/types/querier_test.go#L40-L45
    """

    def __init__(
        self,
        limit: int = 100,
        offset: int = None,
        count_total: bool = True,
        reverse: bool = False,
    ):
        self.limit = limit
        self.offset = offset
        self.count_total = count_total
        self.reverse = reverse

    def build(self, key: bytes = None):
        args = {"limit": self.limit}
        # If key is set --> [It is less efficient than using key. Only one of offset or key should be set]
        if key not in [0x01, None]:
            args["key"] = key
        elif self.offset is not None:
            args["offset"] = self.offset
            args["count_total"] = self.count_total
            args["reverse"] = self.reverse
        return cosmos_pagination_pb2.PageRequest(**args)

    def __str__(self):
        return f"PageRequest(limit={self.limit}, offset={self.offset}, count_total={self.count_total}, reverse={self.reverse})"

    def __repr__(self):
        return f"PageRequest(limit={self.limit}, offset={self.offset}, count_total={self.count_total}, reverse={self.reverse})"



from typing import Any

import grpc
import sentinel_protobuf.sentinel.subscription.v2.querier_pb2 as sentinel_subscription_v2_querier_pb2
import sentinel_protobuf.sentinel.subscription.v2.querier_pb2_grpc as sentinel_subscription_v2_querier_pb2_grpc
import sentinel_protobuf.sentinel.subscription.v2.subscription_pb2 as subscription_pb2

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.types import PageRequest


class SubscriptionQuerier(Querier):
    def __init__(self, channel: grpc.Channel):
        self.__stub = sentinel_subscription_v2_querier_pb2_grpc.QueryServiceStub(
            channel
        )

    def QuerySubscription(self, subscription_id: int) -> Any:
        try:
            r = self.__stub.QuerySubscription(
                sentinel_subscription_v2_querier_pb2.QuerySubscriptionRequest(
                    id=subscription_id
                )
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return self.__ConvertAnyToNodeSubscription(r.subscription.value)

    def QuerySubscriptions(self, pagination: PageRequest = None) -> list:
        subscriptions = self.QueryAll(
            query=self.__stub.QuerySubscriptions,
            request=sentinel_subscription_v2_querier_pb2.QuerySubscriptionsRequest,
            attribute="subscriptions",
            pagination=pagination,
        )
        return [
            self.__ConvertAnyToNodeSubscription(subscription.value)
            for subscription in subscriptions
        ]

    def QueryAllocation(self, address: str, subscription_id: int) -> list:
        try:
            r = self.__stub.QueryAllocation(
                sentinel_subscription_v2_querier_pb2.QueryAllocationRequest(
                    address=address, id=subscription_id
                )
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.allocation

    def QueryAllocations(
        self, subscription_id: int, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QueryAllocations,
            request=sentinel_subscription_v2_querier_pb2.QueryAllocationsRequest,
            attribute="allocations",
            id=subscription_id,
            pagination=pagination,
        )

    def QueryPayout(self, payout_id: int) -> Any:
        try:
            r = self.__stub.QueryPayout(
                sentinel_subscription_v2_querier_pb2.QueryPayoutRequest(id=payout_id)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.payout

    def QueryPayouts(self, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryPayouts,
            request=sentinel_subscription_v2_querier_pb2.QueryPayoutsRequest,
            attribute="payouts",
            pagination=pagination,
        )

    def QueryPayoutsForAccount(
        self, address: str, pagination: PageRequest = None
    ) -> list:
        return self.QueryAll(
            query=self.__stub.QueryPayoutsForAccount,
            request=sentinel_subscription_v2_querier_pb2.QueryPayoutsForAccountRequest,
            attribute="payouts",
            address=address,
            pagination=pagination,
        )

    def QueryPayoutsForNode(self, address: str, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryPayoutsForNode,
            request=sentinel_subscription_v2_querier_pb2.QueryPayoutsForNodeRequest,
            attribute="payouts",
            address=address,
            pagination=pagination,
        )

    def QuerySubscriptionsForAccount(
        self, address: str, pagination: PageRequest = None
    ) -> list:
        subscriptions = self.QueryAll(
            query=self.__stub.QuerySubscriptionsForAccount,
            request=sentinel_subscription_v2_querier_pb2.QuerySubscriptionsForAccountRequest,
            attribute="subscriptions",
            address=address,
            pagination=pagination,
        )
        return [
            self.__ConvertAnyToPlanSubscription(subscription.value)
            for subscription in subscriptions
        ]

    def QuerySubscriptionsForNode(
        self, address: str, pagination: PageRequest = None
    ) -> list:
        subscriptions = self.QueryAll(
            query=self.__stub.QuerySubscriptionsForNode,
            request=sentinel_subscription_v2_querier_pb2.QuerySubscriptionsForNodeRequest,
            attribute="subscriptions",
            address=address,
            pagination=pagination,
        )
        return [
            self.__ConvertAnyToNodeSubscription(subscription.value)
            for subscription in subscriptions
        ]

    def QuerySubscriptionsForPlan(
        self, plan_id: int, pagination: PageRequest = None
    ) -> list:
        subscriptions = self.QueryAll(
            query=self.__stub.QuerySubscriptionsForPlan,
            request=sentinel_subscription_v2_querier_pb2.QuerySubscriptionsForPlanRequest,
            attribute="subscriptions",
            id=plan_id,
            pagination=pagination,
        )
        return [
            self.__ConvertAnyToPlanSubscription(subscription.value)
            for subscription in subscriptions
        ]

    # Node subscriptions are returned by grpc querier in google's 'Any' type and need to be converted into desired protobuf type
    #
    #
    def __ConvertAnyToNodeSubscription(self, any_proto: bytes) -> Any:
        nodesub = subscription_pb2.NodeSubscription()
        nodesub.ParseFromString(any_proto)
        return nodesub

    def __ConvertAnyToPlanSubscription(self, any_proto: bytes) -> Any:
        plansub = subscription_pb2.PlanSubscription()
        plansub.ParseFromString(any_proto)
        return plansub



from typing import Any

import grpc
import sentinel_protobuf.sentinel.deposit.v1.querier_pb2 as sentinel_deposit_v1_querier_pb2
import sentinel_protobuf.sentinel.deposit.v1.querier_pb2_grpc as sentinel_deposit_v1_querier_pb2_grpc

from sentinel_sdk.querier.querier import Querier
from sentinel_sdk.types import PageRequest


class DepositModule(Querier):
    def __init__(self, channel: grpc.Channel):
        self.__stub = sentinel_deposit_v1_querier_pb2_grpc.QueryServiceStub(channel)

    def QueryDeposit(self, address: str) -> Any:
        try:
            r = self.__stub.QueryDeposit(
                sentinel_deposit_v1_querier_pb2.QueryDepositRequest(address=address)
            )
        except grpc._channel._InactiveRpcError as e:
            print(e)
            return None

        return r.deposit

    def QueryDeposits(self, pagination: PageRequest = None) -> list:
        return self.QueryAll(
            query=self.__stub.QueryDeposits,
            request=sentinel_deposit_v1_querier_pb2.QueryDepositsRequest,
            attribute="deposits",
            pagination=pagination,
        )
